local ADDON_NAME, MazeHelper = ...;
local L = MazeHelper.L;

-- Lua API
local tonumber = tonumber;

-- WoW API
local IsInRaid, IsInGroup, GetMinimapZoneText = IsInRaid, IsInGroup, GetMinimapZoneText;

local MEDIA_PATH = 'Interface\\AddOns\\' .. ADDON_NAME .. '\\Media\\';

C_ChatInfo.RegisterAddonMessagePrefix('MAZEHELPER');

local playerNameWithRealm, inInstance, bossKilled, inEncounter, isMinimized;

local FRAME_SIZE = 300;
local X_OFFSET = 2;
local Y_OFFSET = -2;

local MAX_BUTTONS = 8;
local MAX_ACTIVE_BUTTONS = 4;
local NUM_ACTIVE_BUTTONS = 0;

local PASSED_COUNTER = 1;

local MISTCALLER_ENCOUNTER_ID = 2392;

local MazeTexture = {
    TEXTURE = MEDIA_PATH .. 'maze.blp', -- 256x256

    COORDS_WHITE = {
        LEAF_CIRCLE_FILL = {0, 1/4, 0, 1/4},
        LEAF_CIRCLE_NOFILL = {1/4, 2/4, 0, 1/4},
        LEAF_NOCIRCLE_FILL = {0, 1/4, 1/4, 2/4},
        LEAF_NOCIRCLE_NOFILL = {1/4, 2/4, 1/4, 2/4},

        FLOWER_CIRCLE_FILL = {2/4, 3/4, 0, 1/4},
        FLOWER_CIRCLE_NOFILL = {3/4, 1, 0, 1/4},
        FLOWER_NOCIRCLE_FILL = {2/4, 3/4, 1/4, 2/4},
        FLOWER_NOCIRCLE_NOFILL = {3/4, 1, 1/4, 2/4},
    },

    COORDS_COLOR = {
        LEAF_CIRCLE_FILL = {0, 1/4, 2/4, 3/4},
        LEAF_CIRCLE_NOFILL = {1/4, 2/4, 2/4, 3/4},
        LEAF_NOCIRCLE_FILL = {0, 1/4, 3/4, 4/4},
        LEAF_NOCIRCLE_NOFILL = {1/4, 2/4, 3/4, 4/4},

        FLOWER_CIRCLE_FILL = {2/4, 3/4, 2/4, 3/4},
        FLOWER_CIRCLE_NOFILL = {3/4, 4/4, 2/4, 3/4},
        FLOWER_NOCIRCLE_FILL = {2/4, 3/4, 3/4, 4/4},
        FLOWER_NOCIRCLE_NOFILL = {3/4, 1, 3/4, 4/4},
    },
};

local IconsTexture = {
    TEXTURE = MEDIA_PATH .. 'icons32.blp', -- 64x64
    COORDS = {
        MEGAPHONE_WHITE = {0, 1/2, 0, 1/2},
        CROSS_WHITE = {1/2, 2/2, 0, 1/2},
    },
};

local buttons = {}
local buttonsData = {
    [1] = {
        name = L['MAZE_HELPER_LEAF_FULL_CIRCLE'],
        aname = L['MAZE_HELPER_ANNOUNCE_LEAF_FULL_CIRCLE'],
        coords = MazeTexture.COORDS_COLOR.LEAF_CIRCLE_FILL,
        leaf = true,
        flower = false,
        circle = true,
        fill = true,
    },
    [2] = {
        name = L['MAZE_HELPER_LEAF_NOFULL_CIRCLE'],
        aname = L['MAZE_HELPER_ANNOUNCE_LEAF_NOFULL_CIRCLE'],
        coords = MazeTexture.COORDS_COLOR.LEAF_CIRCLE_NOFILL,
        leaf = true,
        flower = false,
        circle = true,
        fill = false,
    },
    [3] = {
        name = L['MAZE_HELPER_FLOWER_FULL_CIRCLE'],
        aname = L['MAZE_HELPER_ANNOUNCE_FLOWER_FULL_CIRCLE'],
        coords = MazeTexture.COORDS_COLOR.FLOWER_CIRCLE_FILL,
        leaf = false,
        flower = true,
        circle = true,
        fill = true,
    },
    [4] = {
        name = L['MAZE_HELPER_FLOWER_NOFULL_CIRCLE'],
        aname = L['MAZE_HELPER_ANNOUNCE_FLOWER_NOFULL_CIRCLE'],
        coords = MazeTexture.COORDS_COLOR.FLOWER_CIRCLE_NOFILL,
        leaf = false,
        flower = true,
        circle = true,
        fill = false,
    },
    [5] = {
        name = L['MAZE_HELPER_LEAF_FULL_NOCIRCLE'],
        aname = L['MAZE_HELPER_ANNOUNCE_LEAF_FULL_NOCIRCLE'],
        coords = MazeTexture.COORDS_COLOR.LEAF_NOCIRCLE_FILL,
        leaf = true,
        flower = false,
        circle = false,
        fill = true,
    },
    [6] = {
        name = L['MAZE_HELPER_LEAF_NOFULL_NOCIRCLE'],
        aname = L['MAZE_HELPER_ANNOUNCE_LEAF_NOFULL_NOCIRCLE'],
        coords = MazeTexture.COORDS_COLOR.LEAF_NOCIRCLE_NOFILL,
        leaf = true,
        flower = false,
        circle = false,
        fill = false,
    },
    [7] = {
        name = L['MAZE_HELPER_FLOWER_FULL_NOCIRCLE'],
        aname = L['MAZE_HELPER_ANNOUNCE_FLOWER_FULL_NOCIRCLE'],
        coords = MazeTexture.COORDS_COLOR.FLOWER_NOCIRCLE_FILL,
        leaf = false,
        flower = true,
        circle = false,
        fill = true,
    },
    [8] = {
        name = L['MAZE_HELPER_FLOWER_NOFULL_NOCIRCLE'],
        aname = L['MAZE_HELPER_ANNOUNCE_FLOWER_NOFULL_NOCIRCLE'],
        coords = MazeTexture.COORDS_COLOR.FLOWER_NOCIRCLE_NOFILL,
        leaf = false,
        flower = true,
        circle = false,
        fill = false,
    },
};

local buttonsMOTSHMap = {
    [1] = 4,
    [2] = 3,
    [3] = 2,
    [4] = 1,
    [5] = 8,
    [6] = 7,
    [7] = 6,
    [8] = 5,
};

local function GetPartyChatType()
    return (not IsInRaid() and IsInGroup(LE_PARTY_CATEGORY_INSTANCE)) and 'INSTANCE_CHAT' or (IsInGroup(LE_PARTY_CATEGORY_HOME) and 'PARTY' or false);
end

local function BetterOnDragStop(frame)
    local point, relativeTo, relativePoint, xOfs, yOfs = frame:GetPoint();

    frame:StopMovingOrSizing();

    frame:ClearAllPoints();
    PixelUtil.SetPoint(frame, point, relativeTo, relativePoint, xOfs, yOfs);
    frame:SetUserPlaced(true);
end

MazeHelper.frame = CreateFrame('Frame', 'ST_Maze_Helper', UIParent);
PixelUtil.SetPoint(MazeHelper.frame, 'CENTER', UIParent, 'CENTER', -FRAME_SIZE, FRAME_SIZE);
PixelUtil.SetSize(MazeHelper.frame, FRAME_SIZE + X_OFFSET * (MAX_ACTIVE_BUTTONS - 1), FRAME_SIZE * 3/4);
MazeHelper.frame:EnableMouse(true);
MazeHelper.frame:SetMovable(true);
MazeHelper.frame:SetClampedToScreen(true);
MazeHelper.frame:RegisterForDrag('LeftButton');
MazeHelper.frame:SetScript('OnDragStart', function(self)
    if self:IsMovable() then
        self:StartMoving();
    end
end);
MazeHelper.frame:SetScript('OnDragStop', BetterOnDragStop);

-- Background
MazeHelper.frame.background = MazeHelper.frame:CreateTexture(nil, 'BACKGROUND');
PixelUtil.SetPoint(MazeHelper.frame.background, 'TOPLEFT', MazeHelper.frame, 'TOPLEFT', -15, 22);
PixelUtil.SetPoint(MazeHelper.frame.background, 'BOTTOMRIGHT', MazeHelper.frame, 'BOTTOMRIGHT', 15, -98);
MazeHelper.frame.background:SetTexture(MEDIA_PATH .. 'background-white.blp');
MazeHelper.frame.background:SetVertexColor(0.05, 0.05, 0.05);
MazeHelper.frame.background:SetAlpha(0.85);

-- Close Button
MazeHelper.frame.CloseButton = CreateFrame('Button', nil, MazeHelper.frame);
PixelUtil.SetPoint(MazeHelper.frame.CloseButton, 'TOPRIGHT', MazeHelper.frame, 'TOPRIGHT', -6, -2);
PixelUtil.SetSize(MazeHelper.frame.CloseButton, 14, 14);
MazeHelper.frame.CloseButton.icon = MazeHelper.frame.CloseButton:CreateTexture(nil, 'OVERLAY');
PixelUtil.SetPoint(MazeHelper.frame.CloseButton.icon, 'CENTER', MazeHelper.frame.CloseButton, 'CENTER', 0, 0);
PixelUtil.SetSize(MazeHelper.frame.CloseButton.icon, 10, 10);
MazeHelper.frame.CloseButton.icon:SetTexture(IconsTexture.TEXTURE);
MazeHelper.frame.CloseButton.icon:SetTexCoord(unpack(IconsTexture.COORDS.CROSS_WHITE));
MazeHelper.frame.CloseButton.icon:SetVertexColor(0.7, 0.7, 0.7, 1);
MazeHelper.frame.CloseButton:SetScript('OnClick', function(self) self:GetParent():Hide(); end);
MazeHelper.frame.CloseButton:SetScript('OnEnter', function(self) self.icon:SetVertexColor(1, 0.85, 0, 1); end);
MazeHelper.frame.CloseButton:SetScript('OnLeave', function(self) self.icon:SetVertexColor(0.7, 0.7, 0.7, 1); end);

-- Minimize Button
MazeHelper.frame.MinButton = CreateFrame('Button', nil, MazeHelper.frame);
PixelUtil.SetPoint(MazeHelper.frame.MinButton, 'RIGHT', MazeHelper.frame.CloseButton, 'LEFT', -8, 0);
PixelUtil.SetSize(MazeHelper.frame.MinButton, 14, 14);
MazeHelper.frame.MinButton.icon = MazeHelper.frame.MinButton:CreateTexture(nil, 'OVERLAY');
PixelUtil.SetPoint(MazeHelper.frame.MinButton.icon, 'BOTTOM', MazeHelper.frame.MinButton, 'BOTTOM', 0, 2);
PixelUtil.SetSize(MazeHelper.frame.MinButton.icon, 10, 2);
MazeHelper.frame.MinButton.icon:SetTexture('Interface\\Buttons\\WHITE8x8');
MazeHelper.frame.MinButton.icon:SetVertexColor(0.7, 0.7, 0.7, 1);
MazeHelper.frame.MinButton:SetScript('OnClick', function(self)
    isMinimized = true;

    PixelUtil.SetHeight(self:GetParent(), 40);
    for i = 1, MAX_BUTTONS do
        buttons[i]:Hide();
    end

    MazeHelper.frame.SolutionText:Hide();
    MazeHelper.frame.ResetButton:Hide();
    MazeHelper.frame.PassedButton:Hide();

    PixelUtil.SetPoint(MazeHelper.frame.background, 'TOPLEFT', MazeHelper.frame, 'TOPLEFT', -15, 0);
    PixelUtil.SetPoint(MazeHelper.frame.background, 'BOTTOMRIGHT', MazeHelper.frame, 'BOTTOMRIGHT', 15, 0);

    PixelUtil.SetPoint(MazeHelper.frame.CloseButton, 'TOPRIGHT', MazeHelper.frame, 'TOPRIGHT', -8, -7);

    MazeHelper.frame.PassedCounter:ClearAllPoints();
    PixelUtil.SetPoint(MazeHelper.frame.PassedCounter, 'LEFT', MazeHelper.frame, 'LEFT', -18, 5);

    MazeHelper.frame.AnnounceButton:SetShown(false);

    MazeHelper.frame.MaxButton:Show();
    self:Hide();
end);
MazeHelper.frame.MinButton:SetScript('OnEnter', function(self) self.icon:SetVertexColor(1, 0.85, 0, 1); end);
MazeHelper.frame.MinButton:SetScript('OnLeave', function(self) self.icon:SetVertexColor(0.7, 0.7, 0.7, 1); end);

-- Maximize Button
MazeHelper.frame.MaxButton = CreateFrame('Button', nil, MazeHelper.frame, 'SharedButtonSmallTemplate');
PixelUtil.SetPoint(MazeHelper.frame.MaxButton, 'TOP', MazeHelper.frame, 'TOP', 0, -4);
MazeHelper.frame.MaxButton:SetText(L['MAZE_HELPER_EXPAND']);
PixelUtil.SetSize(MazeHelper.frame.MaxButton, tonumber(MazeHelper.frame.MaxButton:GetTextWidth()) + 20, 22);
MazeHelper.frame.MaxButton:SetScript('OnClick', function(self)
    isMinimized = false;

    PixelUtil.SetHeight(self:GetParent(), FRAME_SIZE * 3/4);
    for i = 1, MAX_BUTTONS do
        buttons[i]:Show();
    end

    MazeHelper.frame.SolutionText:Show();
    MazeHelper.frame.ResetButton:Show();
    MazeHelper.frame.PassedButton:SetShown(not inEncounter);

    PixelUtil.SetPoint(MazeHelper.frame.background, 'TOPLEFT', MazeHelper.frame, 'TOPLEFT', -15, 22);
    PixelUtil.SetPoint(MazeHelper.frame.background, 'BOTTOMRIGHT', MazeHelper.frame, 'BOTTOMRIGHT', 15, -98);

    PixelUtil.SetPoint(MazeHelper.frame.CloseButton, 'TOPRIGHT', MazeHelper.frame, 'TOPRIGHT', -8, -4);

    MazeHelper.frame.PassedCounter:ClearAllPoints();
    PixelUtil.SetPoint(MazeHelper.frame.PassedCounter, 'BOTTOM', MazeHelper.frame, 'TOP', 0, -32);

    MazeHelper.frame.AnnounceButton:SetShown((MazeHelper.frame.AnnounceButton.solutionAnnounceName and GetPartyChatType()) and true or false);

    MazeHelper.frame.MinButton:Show();
    self:Hide()
end);
MazeHelper.frame.MaxButton:Hide();

-- Solution Text
MazeHelper.frame.SolutionText = MazeHelper.frame:CreateFontString(nil, 'ARTWORK', 'GameFontNormalLarge');
PixelUtil.SetPoint(MazeHelper.frame.SolutionText, 'LEFT', MazeHelper.frame, 'LEFT', 2, -54);
PixelUtil.SetPoint(MazeHelper.frame.SolutionText, 'RIGHT', MazeHelper.frame, 'RIGHT', -2, 0);
MazeHelper.frame.SolutionText:SetShadowColor(0.15, 0.15, 0.15);
MazeHelper.frame.SolutionText:SetText(L['MAZE_HELPER_CHOOSE_SYMBOLS_4']);

local function ResetAll()
    for i = 1, MAX_BUTTONS do
        MazeHelper:SetUnactiveButton(buttons[i]);

        buttons[i].state = false;
        buttons[i].sender = nil;
    end

    NUM_ACTIVE_BUTTONS = 0;
    MazeHelper.frame.SolutionText:SetText(L['MAZE_HELPER_CHOOSE_SYMBOLS_4']);
    MazeHelper.frame.PassedButton:SetEnabled(false);
    MazeHelper.frame.AnnounceButton:SetShown(false);
    MazeHelper.frame.AnnounceButton.solutionAnnounceName = nil;

    MazeHelper.frame.ResetButton:SetEnabled(false);
end

-- Reset Button
MazeHelper.frame.ResetButton = CreateFrame('Button', nil, MazeHelper.frame, 'SharedButtonSmallTemplate');
PixelUtil.SetPoint(MazeHelper.frame.ResetButton, 'TOP', MazeHelper.frame.SolutionText, 'BOTTOM', 0, -8);
MazeHelper.frame.ResetButton:SetText(L['MAZE_HELPER_RESET']);
PixelUtil.SetSize(MazeHelper.frame.ResetButton, tonumber(MazeHelper.frame.ResetButton:GetTextWidth()) + 20, 22);
MazeHelper.frame.ResetButton:SetScript('OnClick', function()
    if NUM_ACTIVE_BUTTONS == 0 then
        return;
    end

    MazeHelper:SendResetCommand();
    ResetAll();
end);
MazeHelper.frame.ResetButton:SetEnabled(false);

-- Passed Button
MazeHelper.frame.PassedButton = CreateFrame('Button', nil, MazeHelper.frame, 'SharedButtonSmallTemplate');
PixelUtil.SetPoint(MazeHelper.frame.PassedButton, 'RIGHT', MazeHelper.frame.ResetButton, 'LEFT', -8, 0);
MazeHelper.frame.PassedButton:SetText(L['MAZE_HELPER_PASSED']);
PixelUtil.SetSize(MazeHelper.frame.PassedButton, tonumber(MazeHelper.frame.PassedButton:GetTextWidth()) + 20, 22);
MazeHelper.frame.PassedButton:SetScript('OnClick', function()
    PASSED_COUNTER = PASSED_COUNTER + 1;

    MazeHelper.frame.PassedCounter.Text:SetText(PASSED_COUNTER);
    MazeHelper:SendPassedCommand(PASSED_COUNTER);

    ResetAll();
end);
MazeHelper.frame.PassedButton:SetEnabled(false);

-- Reposition Reset Button
PixelUtil.SetPoint(MazeHelper.frame.ResetButton, 'TOP', MazeHelper.frame.SolutionText, 'BOTTOM', tonumber(MazeHelper.frame.PassedButton:GetTextWidth()) - 8, -8);

-- Passed Counter Text
MazeHelper.frame.PassedCounter = CreateFrame('Frame', nil, MazeHelper.frame);
PixelUtil.SetPoint(MazeHelper.frame.PassedCounter, 'BOTTOM', MazeHelper.frame, 'TOP', 0, -32);
PixelUtil.SetSize(MazeHelper.frame.PassedCounter, 64, 64);

MazeHelper.frame.PassedCounter.Background = MazeHelper.frame.PassedCounter:CreateTexture(nil, 'BACKGROUND');
MazeHelper.frame.PassedCounter.Background:SetAllPoints();
MazeHelper.frame.PassedCounter.Background:SetTexture(MEDIA_PATH .. 'bbh.blp');

MazeHelper.frame.PassedCounter.Text = MazeHelper.frame.PassedCounter:CreateFontString(nil, 'ARTWORK', 'GameFontNormalShadowHuge2');
PixelUtil.SetPoint(MazeHelper.frame.PassedCounter.Text, 'CENTER', MazeHelper.frame.PassedCounter, 'CENTER', 0, 0);
PixelUtil.SetSize(MazeHelper.frame.PassedCounter.Text, 0, 20);
MazeHelper.frame.PassedCounter.Text:SetShadowColor(0.15, 0.15, 0.15);
MazeHelper.frame.PassedCounter.Text:SetText(PASSED_COUNTER);
MazeHelper.frame.PassedCounter.Text:SetJustifyH('CENTER');

-- Announce Button
MazeHelper.frame.AnnounceButton = CreateFrame('Button', nil, MazeHelper.frame);
PixelUtil.SetPoint(MazeHelper.frame.AnnounceButton, 'TOPLEFT', MazeHelper.frame, 'TOPLEFT', 2, 4);
PixelUtil.SetSize(MazeHelper.frame.AnnounceButton, 24, 24);
MazeHelper.frame.AnnounceButton.icon = MazeHelper.frame.AnnounceButton:CreateTexture(nil, 'OVERLAY');
PixelUtil.SetPoint(MazeHelper.frame.AnnounceButton.icon, 'CENTER', MazeHelper.frame.AnnounceButton, 'CENTER', 0, 0);
PixelUtil.SetSize(MazeHelper.frame.AnnounceButton.icon, 18, 18);
MazeHelper.frame.AnnounceButton.icon:SetTexture(IconsTexture.TEXTURE);
MazeHelper.frame.AnnounceButton.icon:SetTexCoord(unpack(IconsTexture.COORDS.MEGAPHONE_WHITE));
MazeHelper.frame.AnnounceButton.icon:SetVertexColor(0.7, 0.7, 0.7, 1);
MazeHelper.frame.AnnounceButton:SetScript('OnClick', function(self)
    if not self.solutionAnnounceName then
        return;
    end

    local partyChatType = GetPartyChatType();
    if partyChatType then
        SendChatMessage(string.format(L['MAZE_HELPER_ANNOUNCE_SOLUTION'], self.solutionAnnounceName), partyChatType);
    end

    self:SetShown(false);
end);
MazeHelper.frame.AnnounceButton:SetScript('OnEnter',function(self) self.icon:SetVertexColor(1, 0.85, 0, 1); end);
MazeHelper.frame.AnnounceButton:SetScript('OnLeave',function(self) self.icon:SetVertexColor(0.7, 0.7, 0.7, 1); end);
MazeHelper.frame.AnnounceButton:SetShown(false);

function MazeHelper:CreateButton(index)
    local button = CreateFrame('Button', nil, self.frame, 'BackdropTemplate');

    if index == 1 then
        PixelUtil.SetPoint(button, 'TOPLEFT', self.frame, 'TOPLEFT', 20, -20);
    elseif index == 5 then
        PixelUtil.SetPoint(button, 'TOPLEFT', buttons[1], 'BOTTOMLEFT', 0, Y_OFFSET);
    else
        PixelUtil.SetPoint(button, 'LEFT', buttons[index - 1], 'RIGHT', X_OFFSET, 0);
    end

    PixelUtil.SetSize(button, 64, 64);

    button.Icon = button:CreateTexture(nil, 'ARTWORK');
    PixelUtil.SetPoint(button.Icon, 'TOPLEFT', button, 'TOPLEFT', 4, -4);
    PixelUtil.SetPoint(button.Icon, 'BOTTOMRIGHT', button, 'BOTTOMRIGHT', -4, 4);
    button.Icon:SetTexture(MazeTexture.TEXTURE);
    button.Icon:SetTexCoord(unpack(buttonsData[index].coords));

    button:SetBackdrop({
        insets = {top = 1, left = 1, bottom = 1, right = 1},
        edgeFile = 'Interface\\Buttons\\WHITE8x8',
		edgeSize = 2,
    });

    MazeHelper:SetUnactiveButton(button);

    button.id    = index;
    button.data  = buttonsData[index];
    button.state = false;

    button:RegisterForClicks('LeftButtonUp', 'RightButtonUp');

    button:SetScript('OnClick', function(self, button)
        if button == 'LeftButton' then
            if self.state or NUM_ACTIVE_BUTTONS == MAX_ACTIVE_BUTTONS then
                return;
            end

            MazeHelper.frame.ResetButton:SetEnabled(true);

            NUM_ACTIVE_BUTTONS = math.min(MAX_ACTIVE_BUTTONS, NUM_ACTIVE_BUTTONS + 1);
            self.state = true;
            MazeHelper:SetActiveButton(self);
            MazeHelper.frame.SolutionText:SetText(L['MAZE_HELPER_CHOOSE_SYMBOLS_' .. (MAX_ACTIVE_BUTTONS - NUM_ACTIVE_BUTTONS)]);

            MazeHelper:SendButtonID(self.id, 'ACTIVE');

            if NUM_ACTIVE_BUTTONS == MAX_ACTIVE_BUTTONS then
                MazeHelper:GetSolution();
            end
        elseif button == 'RightButton' then
            if not self.state then
                return;
            end

            if self.state then
                NUM_ACTIVE_BUTTONS = math.max(0, NUM_ACTIVE_BUTTONS - 1);
                self.state = false;
                self.sender = nil;
            end

            if NUM_ACTIVE_BUTTONS < MAX_ACTIVE_BUTTONS then
                MazeHelper.frame.SolutionText:SetText(L['MAZE_HELPER_CHOOSE_SYMBOLS_' .. (MAX_ACTIVE_BUTTONS - NUM_ACTIVE_BUTTONS)]);

                for i = 1, MAX_BUTTONS do
                    if buttons[i].state then
                        MazeHelper:SetActiveButton(buttons[i]);
                    end
                end

                MazeHelper.frame.PassedButton:SetEnabled(false);
                MazeHelper.frame.AnnounceButton:SetShown(false);
                MazeHelper.frame.AnnounceButton.solutionAnnounceName = nil;
            end

            if NUM_ACTIVE_BUTTONS == 0 then
                MazeHelper.frame.ResetButton:SetEnabled(false);
            end

            MazeHelper:SetUnactiveButton(self);
            MazeHelper:SendButtonID(self.id, 'UNACTIVE');
        end
    end);

    button:SetScript('OnEnter', function(self)
        if not self.sender then
            return;
        end

        GameTooltip:SetOwner(self, 'ANCHOR_RIGHT');
        GameTooltip:AddLine(self.state and string.format(L['MAZE_HELPER_SENDED_BY'], self.sender) or string.format(L['MAZE_HELPER_CLEARED_BY'], self.sender), 1, 0.85, 0, true);
        GameTooltip:Show();
    end);
    button:SetScript('OnLeave', GameTooltip_Hide);

    button:RegisterForDrag('LeftButton');
    button:SetScript('OnDragStart', function(self)
        if self:GetParent():IsMovable() then
            self:GetParent():StartMoving();
        end
    end);
    button:SetScript('OnDragStop', function(self)
        BetterOnDragStop(self:GetParent());
    end);

    table.insert(buttons, index, button); -- index for just to be sure
end

function MazeHelper:CreateButtons()
    for i = 1, MAX_BUTTONS do
        MazeHelper:CreateButton(i);
    end
end

function MazeHelper:SetUnactiveButton(button)
    button:SetBackdropBorderColor(0, 0, 0, 0);
end

function MazeHelper:SetActiveButton(button)
    button:SetBackdropBorderColor(0.4, 0.52, 0.95, 1);
end

function MazeHelper:SetReceivedButton(button)
    button:SetBackdropBorderColor(0.9, 1, 0.1, 1);
end

function MazeHelper:SetSolutionButton(button)
    button:SetBackdropBorderColor(0.2, 0.8, 0.4, 1);
end

-- DON'T LOOK AT THIS SHIT PLEEEEASE :( I WAS DRUNK, BUT IT STILL WORKS SOMEHOW
function MazeHelper:GetSolution()
    local circleSum, flowerSum, leafSum, fillSum = 0, 0, 0, 0;
    for i = 1, MAX_BUTTONS do
        if buttons[i].state then
            if buttons[i].data.circle then
                circleSum = circleSum + 1;
            end

            if buttons[i].data.flower then
                flowerSum = flowerSum + 1;
            end

            if buttons[i].data.leaf then
                leafSum = leafSum + 1;
            end

            if buttons[i].data.fill then
                fillSum = fillSum + 1;
            end
        end
    end

    local fill, flower, leaf, circle;
    if fillSum == 3 then
        fill = false;
    elseif fillSum == 1 then
        fill = true;
    end

    if flowerSum == 3 then
        flower = false
    elseif flowerSum == 1 then
        flower = true;
    end

    if leafSum == 3 then
        leaf = false;
    elseif leafSum == 1 then
        leaf = true;
    end

    if circleSum == 3 then
        circle = false;
    elseif circleSum == 1 then
        circle = true;
    end

    local solutionName, solutionAnnounceName, solutionButtonID;
    for i = 1, MAX_BUTTONS do
        if buttons[i].state then
            if buttons[i].data.fill == fill then
                solutionName = buttons[i].data.name;
                solutionAnnounceName = buttons[i].data.aname;
                solutionButtonID = i;
            end

            if buttons[i].data.leaf == leaf then
                solutionName = buttons[i].data.name;
                solutionAnnounceName = buttons[i].data.aname;
                solutionButtonID = i;
            end

            if buttons[i].data.flower == flower then
                solutionName = buttons[i].data.name;
                solutionAnnounceName = buttons[i].data.aname;
                solutionButtonID = i;
            end

            if buttons[i].data.circle == circle then
                solutionName = buttons[i].data.name;
                solutionAnnounceName = buttons[i].data.aname;
                solutionButtonID = i;
            end
        end
    end

    if solutionButtonID then
        MazeHelper:SetSolutionButton(buttons[solutionButtonID]);
    end

    MazeHelper.frame.AnnounceButton.solutionAnnounceName = solutionAnnounceName or nil;
    MazeHelper.frame.AnnounceButton:SetShown((not isMinimized and solutionAnnounceName and GetPartyChatType()) and true or false);

    MazeHelper.frame.PassedButton:SetEnabled(solutionButtonID and true or false);
    MazeHelper.frame.SolutionText:SetText(solutionName and string.format(L['MAZE_HELPER_SOLUTION'], solutionName) or L['MAZE_HELPER_SOLUTION_NA']);
end

function MazeHelper:SendResetCommand()
    local partyChatType = GetPartyChatType();
    if partyChatType then
        C_ChatInfo.SendAddonMessage('MAZEHELPER', 'SendReset', partyChatType);

        -- Mists of Tirna Scithe Helper addon compatibility
        C_ChatInfo.SendAddonMessage('MOTSH', 'reset', partyChatType);
    end
end

function MazeHelper:SendPassedCommand(step)
    local partyChatType = GetPartyChatType();
    if partyChatType then
        local data = 'SendPassed|'.. step;
        C_ChatInfo.SendAddonMessage('MAZEHELPER', data, partyChatType);

        -- Mists of Tirna Scithe Helper addon compatibility
        C_ChatInfo.SendAddonMessage('MOTSH', 'reset', partyChatType);
    end
end

function MazeHelper:SendButtonID(buttonID, mode)
    local partyChatType = GetPartyChatType();
    if partyChatType then
        local data = 'SendButtonID|' .. buttonID .. '|' .. mode;
        C_ChatInfo.SendAddonMessage('MAZEHELPER', data, partyChatType);

        -- Mists of Tirna Scithe Helper addon compatibility
        C_ChatInfo.SendAddonMessage('MOTSH', buttonsMOTSHMap[buttonID], partyChatType);
    end
end

function MazeHelper:ReceiveResetCommand()
    ResetAll();
end

function MazeHelper:ReceivePassedCommand(step)
    PASSED_COUNTER = step;

    MazeHelper.frame.PassedCounter.Text:SetText(PASSED_COUNTER);

    ResetAll();
end

function MazeHelper:ReceiveActiveButtonID(buttonID, sender)
    if not buttons[buttonID] then
        return;
    end

    if buttons[buttonID].state or NUM_ACTIVE_BUTTONS == MAX_ACTIVE_BUTTONS then
        return;
    end

    NUM_ACTIVE_BUTTONS = math.min(MAX_ACTIVE_BUTTONS, NUM_ACTIVE_BUTTONS + 1);
    buttons[buttonID].state = true;
    buttons[buttonID].sender = sender;

    MazeHelper:SetReceivedButton(buttons[buttonID]);
    MazeHelper.frame.SolutionText:SetText(L['MAZE_HELPER_CHOOSE_SYMBOLS_' .. (MAX_ACTIVE_BUTTONS - NUM_ACTIVE_BUTTONS)]);

    MazeHelper.frame.ResetButton:SetEnabled(true);

    if NUM_ACTIVE_BUTTONS == MAX_ACTIVE_BUTTONS then
        MazeHelper:GetSolution();
    end
end

function MazeHelper:ReceiveUnactiveButtonID(buttonID, sender)
    if not buttons[buttonID] then
        return;
    end

    if not buttons[buttonID].state then
        return;
    end

    NUM_ACTIVE_BUTTONS = math.max(0, NUM_ACTIVE_BUTTONS - 1);
    buttons[buttonID].state = false;
    buttons[buttonID].sender = sender;

    if NUM_ACTIVE_BUTTONS < MAX_ACTIVE_BUTTONS then
        MazeHelper.frame.SolutionText:SetText(L['MAZE_HELPER_CHOOSE_SYMBOLS_' .. (MAX_ACTIVE_BUTTONS - NUM_ACTIVE_BUTTONS)]);

        for i = 1, MAX_BUTTONS do
            if buttons[i].state then
                MazeHelper:SetActiveButton(buttons[i]);
            end
        end

        MazeHelper.frame.PassedButton:SetEnabled(false);
        MazeHelper.frame.AnnounceButton:SetShown(false);
        MazeHelper.frame.AnnounceButton.solutionAnnounceName = nil;
    end

    if NUM_ACTIVE_BUTTONS == 0 then
        MazeHelper.frame.ResetButton:SetEnabled(false);
    end

    MazeHelper:SetUnactiveButton(buttons[buttonID]);
end

local function UpdateShown()
    MazeHelper.frame:SetShown((not bossKilled and inInstance and GetMinimapZoneText() == L['MAZE_HELPER_ZONE_NAME']));
end

MazeHelper.frame:RegisterEvent('ADDON_LOADED');
MazeHelper.frame:SetScript('OnEvent', function(self, event, ...)
    if self[event] then
        return self[event](self, ...);
    end
end);

local function UpdateData(frame) -- Not good name, i know...
    playerNameWithRealm = UnitName('player') .. '-' .. (GetNormalizedRealmName() or string.gsub(GetRealmName(), '[%s%-]', '')); -- Just for sure...
    inInstance = IsInInstance();
    inEncounter = not bossKilled and UnitExists('boss1');
    PASSED_COUNTER = 1;

    if inInstance then
        frame:RegisterEvent('ZONE_CHANGED');
        frame:RegisterEvent('ZONE_CHANGED_INDOORS');
        frame:RegisterEvent('ZONE_CHANGED_NEW_AREA');
        frame:RegisterEvent('ENCOUNTER_START');
        frame:RegisterEvent('ENCOUNTER_END');
        frame:RegisterEvent('BOSS_KILL');
    else
        frame:UnregisterEvent('ZONE_CHANGED');
        frame:UnregisterEvent('ZONE_CHANGED_INDOORS');
        frame:UnregisterEvent('ZONE_CHANGED_NEW_AREA');
        frame:UnregisterEvent('ENCOUNTER_START');
        frame:UnregisterEvent('ENCOUNTER_END');
        frame:UnregisterEvent('BOSS_KILL');
    end

    if inEncounter then
        PixelUtil.SetPoint(MazeHelper.frame.ResetButton, 'TOP', MazeHelper.frame.SolutionText, 'BOTTOM', 0, -8);
        MazeHelper.frame.PassedButton:Hide();
        MazeHelper.frame.PassedCounter:Hide();
    else
        if not bossKilled then
            PixelUtil.SetPoint(MazeHelper.frame.ResetButton, 'TOP', MazeHelper.frame.SolutionText, 'BOTTOM', tonumber(MazeHelper.frame.PassedButton:GetTextWidth()) - 8, -8);
            MazeHelper.frame.PassedButton:Show();
            MazeHelper.frame.PassedCounter:Show();
        end
    end

    UpdateShown();
end

function MazeHelper.frame:PLAYER_LOGIN()
    UpdateData(self);
end

function MazeHelper.frame:PLAYER_ENTERING_WORLD()
    UpdateData(self);
end

function MazeHelper.frame:ZONE_CHANGED()
    UpdateShown();
end

function MazeHelper.frame:ZONE_CHANGED_INDOORS()
    UpdateShown();
end

function MazeHelper.frame:ZONE_CHANGED_NEW_AREA()
    UpdateShown();
end

function MazeHelper.frame:BOSS_KILL(encounterID)
    if encounterID == MISTCALLER_ENCOUNTER_ID then
        bossKilled = true;
        UpdateShown();
    end
end

function MazeHelper.frame:ENCOUNTER_START(encounterID)
    if encounterID == MISTCALLER_ENCOUNTER_ID then
        inEncounter = true;

        ResetAll();

        -- Mists of Tirna Scithe Helper addon compatibility
        local partyChatType = GetPartyChatType();
        if partyChatType then
            C_ChatInfo.SendAddonMessage('MOTSH', 'reset', partyChatType);
        end

        bossKilled = false;
        UpdateShown();

        PixelUtil.SetPoint(MazeHelper.frame.ResetButton, 'TOP', MazeHelper.frame.SolutionText, 'BOTTOM', 0, -8);
        MazeHelper.frame.PassedButton:Hide();
        MazeHelper.frame.PassedCounter:Hide();
    end
end

function MazeHelper.frame:ENCOUNTER_END(encounterID, _, _, _, success)
    if encounterID == MISTCALLER_ENCOUNTER_ID then
        inEncounter = false;

        ResetAll();

        -- Mists of Tirna Scithe Helper addon compatibility
        local partyChatType = GetPartyChatType();
        if partyChatType then
            C_ChatInfo.SendAddonMessage('MOTSH', 'reset', partyChatType);
        end

        bossKilled = success;
        UpdateShown();
    end
end

function MazeHelper.frame:CHAT_MSG_ADDON(prefix, message, _, sender)
    if sender == playerNameWithRealm then
        return;
    end

    if prefix == 'MAZEHELPER' then
        local p, buttonID, mode = strsplit('|', message);
        if p == 'SendButtonID'  then
            if mode == 'ACTIVE' then
                MazeHelper:ReceiveActiveButtonID(tonumber(buttonID), sender);
            elseif mode == 'UNACTIVE' then
                MazeHelper:ReceiveUnactiveButtonID(tonumber(buttonID), sender);
            end
        elseif p == 'SendPassed' then
            MazeHelper:ReceivePassedCommand(tonumber(buttonID));
            print(string.format(L['MAZE_HELPER_PASSED_PLAYER'], sender));
        elseif p == 'SendReset' then
            MazeHelper:ReceiveResetCommand();
            print(string.format(L['MAZE_HELPER_RESETED_PLAYER'], sender));
        end
	end
end

function MazeHelper.frame:ADDON_LOADED(addonName)
    if addonName ~= ADDON_NAME then
        return;
    end

    self:UnregisterEvent('ADDON_LOADED');

    MazeHelper:CreateButtons();

    self:RegisterEvent('PLAYER_LOGIN');
    self:RegisterEvent('PLAYER_ENTERING_WORLD');
    self:RegisterEvent('CHAT_MSG_ADDON');

    _G['SLASH_MAZEHELPER1'] = '/mh';
    SlashCmdList['MAZEHELPER'] = function()
        MazeHelper.frame:SetShown(not MazeHelper.frame:IsShown());
    end
end